﻿/*** Name:      NetDataContractStoreBase.cs
 *** Created:   9/23/2008, by: Hilkens, RGM
 *** Modified:  12/2/2008, by: Hilkens, RGM
 *** Source:    Hulpklassen voor serialisatie van objecten.
 *** Copyright: (c) Gr2 Dev 2008
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.Serialization;
using System.Xml.Serialization;
using System.IO;
using System.Xml;

namespace syscore
{
    /// <summary>
    /// Basis klasse voor alle store klassen.
    /// </summary>
    /// <typeparam name="T">Type van data.</typeparam>
    public abstract class AbstractStoreBase<T> where T : new()
    {
        /// <summary>
        /// Bestandspad.
        /// </summary>
        public string Path = "";

        /// <summary>
        /// Copyright/commentaar in bestand.
        /// </summary>
        public string Comment = "";

        /// <summary>
        /// Daadwerkelijke data om op te slaan.
        /// </summary>
        public T InnerData;

        /// <summary>
        /// Slaat klasse in een bestand op.
        /// </summary>
        public abstract void SaveSettings();

        /// <summary>
        /// Laadt klasse van een bestand.
        /// </summary>
        /// <returns>Een klasse met de uitgelezen data.</returns>
        public abstract T LoadSettings();
    }

    /// <summary>
    /// Windows Communication Foundation implementatie van AbstractStoreBase.
    /// </summary>
    /// <typeparam name="T">Type van data.</typeparam>
    public class NetDataContractStoreBase<T> : AbstractStoreBase<T> where T : new()
    {
        /// <summary>
        /// <i>true</i> als .NET Serialisatie gebruikt moet worden, anderzijds <i>false</i>.
        /// </summary>
        public bool UseNetSerializer = true;

        /// <summary>
        /// Slaat klasse in een XML-bestand op.
        /// </summary>
        public override void SaveSettings()
        {
            //Init.
            FileStream filestr = new FileStream(Path, FileMode.Create, FileAccess.Write, FileShare.Read);
            XmlTextWriter xmlWr = new XmlTextWriter(filestr, new System.Text.UTF8Encoding());
            XmlObjectSerializer formatter;
            if (UseNetSerializer) formatter = new NetDataContractSerializer(); else formatter = new DataContractSerializer(InnerData.GetType());

            xmlWr.Formatting = Formatting.Indented;
            xmlWr.WriteComment(Comment);
            formatter.WriteObject(xmlWr, InnerData);
            xmlWr.Flush();
            filestr.Close();
        }

        /// <summary>
        /// Laadt klasse van een XML-bestand.
        /// </summary>
        /// <returns>Een klasse met de uitgelezen data.</returns>
        public override T LoadSettings()
        {
            if (!File.Exists(Path))
            {
                //Maak nieuw bestand.
                InnerData = new T();
                SaveSettings();
                return InnerData;
            }
            else
            {
                //Laad gegevens.
                FileStream filestr = new FileStream(Path, FileMode.Open, FileAccess.Read);
                XmlObjectSerializer formatter;
                if (UseNetSerializer) formatter = new NetDataContractSerializer(); else formatter = new DataContractSerializer(InnerData.GetType());

                InnerData = (T)(formatter.ReadObject(filestr));
                filestr.Close();
                return InnerData;
            }
        }
    }

    /// <summary>
    /// XML Implementatie van AbstractStoreBase.
    /// </summary>
    /// <typeparam name="T">Type van data.</typeparam>
    public class XMLStoreBase<T> : AbstractStoreBase<T> where T : new()
    {
        /// <summary>
        /// Slaat klasse in een XML-bestand op.
        /// </summary>
        public override void SaveSettings()
        {
            //Init.
            FileStream filestr = new FileStream(Path, FileMode.Create, FileAccess.Write, FileShare.Read);
            XmlTextWriter xmlWr = new XmlTextWriter(filestr, new System.Text.UTF8Encoding());
            XmlSerializer xmlSer = new XmlSerializer(typeof(T));

            xmlWr.Formatting = Formatting.Indented;
            xmlWr.WriteComment(Comment);
            xmlSer.Serialize(xmlWr, InnerData);
            filestr.Close();
        }

        /// <summary>
        /// Laadt klasse van een XML-bestand.
        /// </summary>
        /// <returns>Een klasse met de uitgelezen data..</returns>
        public override T LoadSettings()
        {
            if (!File.Exists(Path))
            {
                //Maak nieuw bestand.
                InnerData = new T();
                SaveSettings();
                return InnerData;
            }
            else
            {
                //Laad gegevens.
                FileStream filestr = new FileStream(Path, FileMode.Open, FileAccess.Read);
                XmlSerializer xmlSer = new XmlSerializer(typeof(T));
                InnerData = (T)(xmlSer.Deserialize(filestr));
                filestr.Close();
                return InnerData;
            }
        }
    }
}
