﻿/*** Name:      ConfigurationFile.cs
 *** Created:   11/6/2008, by: Hilkens, RGM
 *** Modified:  12/14/2008, by: Hilkens, RGM
 *** Source:    Representeert een configuratiebestand.
 *** Copyright: (c) Gr2 Dev 2008
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Etaris.Core
{
    /// <summary>
    /// Representeert een configuratiebestand welke speelstuktemplates en speelvelden definieert.
    /// </summary>
    /// <remarks>
    /// In deze klasse refereert 'template' altijd naar een speelstuktemplate.
    /// </remarks>
    public class ConfigurationFile
    {
        #region Velden
        /// <summary>
        /// Breedte van een speelveld in blokken.
        /// </summary>
        int width;

        /// <summary>
        /// Hoogte van een speelveld in blokken.
        /// </summary>
        int height;

        /// <summary>
        /// Tijdelijke lijst met blokken van een speelstuktemplate dat ingelezen wordt.
        /// </summary>
        List<BlockTemplate> blocks;

        /// <summary>
        /// De ingelezen speelstuktemplates van dit configuratiebestand.
        /// </summary>
        List<PieceTemplate> pieces = new List<PieceTemplate>();

        /// <summary>
        /// Maximale toegestane breedte in blokken van een speelveld.
        /// </summary>
        const int maximumFieldWidth = 100;

        /// <summary>
        /// Maximale toegestane hoogte in blokken van een speelveld.
        /// </summary>
        const int maximumFieldHeight = 100;
        #endregion

        /// <summary>
        /// Constructor. Initialiseert de klasse.
        /// </summary>
        private ConfigurationFile()
        { }

        /// <summary>
        /// Breedte van een speelveld in blokken.
        /// </summary>
        public int PlayingFieldWidth
        {
            get { return width; }
        }

        /// <summary>
        /// Hoogte van een speelveld in blokken.
        /// </summary>
        public int PlayingFieldHeight
        {
            get { return height; }
        }

        /// <summary>
        /// Aantal speelstuktemplates in configuratiebestand.
        /// </summary>
        public int PieceTemplatesCount
        {
            get { return pieces.Count; }
        }

        /// <summary>
        /// Laadt een configuratiebestand.
        /// </summary>
        /// <param name="path">Pad van het configuratiebestand.</param>
        /// <returns>Een geldig configuratiebestand.</returns>
        public static ConfigurationFile LoadConfigurationFile(string path)
        {
            //Init.
            ConfigurationFile cf = new ConfigurationFile();
            StreamReader sr = null;
            Exception exception = null;
            bool hasError = false;  //Geeft een fout aan in ProcessConfigurationFile.
            string msg;             //Eventuele exceptie van CheckValidity.

            //Lees configuratiebestand uit.
            try
            {
                sr = new StreamReader(path);
                cf.ProcessConfigurationFile(sr);
            }
            catch (Exception ex)
            {
                exception = ex; hasError = true;
            }
            finally //Zorgt ervoor dat een geopend bestand altijd gesloten wordt.
            { if (sr != null) sr.Close(); }

            //Controleer validiteit.
            if (hasError) throw exception;
            if (!cf.CheckValidity(out msg))
                throw new FormatException(msg);

            //Finish.
            return cf;
        }

        /// <summary>
        /// Geeft de breedte van het breedste en de hoogte van het hoogste speelstuktemplate.
        /// </summary>
        /// <param name="width">Breedte in blokken.</param>
        /// <param name="height">Hoogte in blokken.</param>
        public void GetMaximumPieceDimensions(out int width, out int height)
        {
            if (pieces.Count == 0) { width = 0; height = 0; return; }

            width = pieces[0].Width;
            height = pieces[0].Height;

            for (int i = 0; i < pieces.Count; i++)
            {
                if (pieces[i].Width > width) width = pieces[i].Width;
                if (pieces[i].Height > height) height = pieces[i].Height;
            }
        }

        /// <summary>
        /// Geeft een speelstuktemplate op basis van een index.
        /// </summary>
        /// <param name="index">Index van het speelstuktemplate.</param>
        /// <returns>Een speelstuktemplate.</returns>
        public PieceTemplate GetPieceTemplate(int index)
        {
            return pieces[index];
        }

        /// <summary>
        /// Controleert de geldigheid van ingelezen data.
        /// </summary>
        /// <param name="message">Een string die een foutmelding kan bevatten.</param>
        /// <returns><i>true</i> als de ingelezen data geldig is, anderzijds <i>false</i>.</returns>
        private bool CheckValidity(out string message)
        {
            //Init.
            message = "";
            int i;
            int pieceWidth = 0;
            int pieceHeight = 0;

            //Doe controles op invoer van het configuratiebestand.
            if (width <= 0 || width > maximumFieldWidth)
            {
                message = "Incorrect field width: ";
                if (width <= 0)
                    message += "The width should be positive.\n";
                else
                    message += "The width should be at most 100.\n";
            }

            if (height <= 0 || height > maximumFieldHeight)
            {
                message += "Incorrect field height: ";
                if (height <= 0)
                    message += "The height should be positive.\n";
                else
                    message += "The height should be at most 100.\n";
            }

            //Controleer op minimum aantal speelstuktemplates.
            if (pieces.Count == 0)
                message += "At least one piece template should be present.\n";

            //Controleeer alle speelstuktemplates.
            for (i = 0; i < pieces.Count; i++)
            {
                if (!pieces[i].ValidatePiece())
                    message += string.Format("Piece template {0} is either too big or inconsistent.\n", i + 1); //i+1, omdat tellen bij 1 begint!
            }

            //Controleert de minimale grootte van het speelveld.
            GetMaximumPieceDimensions(out pieceWidth, out pieceHeight);
            if (width < pieceWidth || height < pieceHeight)
                message += "The playingfield is smaller than the largest piece.\n";

            //Finish.
            return (message == "");
        }

        /// <summary>
        /// Leest het geopende configuratiebestand uit.
        /// </summary>
        /// <param name="streamReader">Een geldige StreamReader voor het configuratiebestand.</param>
        private void ProcessConfigurationFile(StreamReader streamReader)
        {
            //Init.
            StreamReader sr = streamReader;
            pieces.Clear();
            string l;
            bool inPiece = false; //Geeft aan of de huidige regel een speelstukdefinitie bevat.
            int row = 0;

            //Breedte en hoogte van het speelveld uitlezen.
            width = int.Parse(sr.ReadLine());
            height = int.Parse(sr.ReadLine());

            //Overige regels uitlezen.
            while ((l = sr.ReadLine()) != null)
            {
                if (inPiece && l == "")
                    inPiece = false; //Einde van een speelstuk.

                else if (!inPiece && l != "")
                {
                    //Maak een nieuw speelstuk.
                    pieces.Add(new PieceTemplate());
                    blocks = pieces[pieces.Count - 1].Blocks;
                    inPiece = true;
                    row = 0;
                }

                //Een rij van een speelstuktemplate verwerken.
                if (inPiece)
                {
                    ProcessLine(l, row);
                    row += 1;
                }
            }
        }

        /// <summary>
        /// Verwerkt een speelstukdefinitieregel van het configuratiebestand.
        /// </summary>
        /// <param name="l">Een regel.</param>
        /// <param name="row">Rij van het speelstuk.</param>
        private void ProcessLine(string l, int row)
        {
            //Init.
            int i; //Kolom van een speelstuk.
            char[] chars = l.ToCharArray();

            for (i = 0; i < chars.Length; i++)
            {
                if (chars[i] == '*')
                    blocks.Add(new BlockTemplate(i, row));
                else if (chars[i] != '-')
                    throw new FormatException("Unsupported character in file.");
            }
        }
    }
}
