﻿/*** Name:      MusicPlayer.cs
 *** Created:   11/6/2008, by: Hilkens, RGM
 *** Modified:  12/14/2008, by: Hilkens, RGM
 *** Source:    Implementeert een simpele muziekspeler.
 *** Copyright: (c) Gr2 Dev 2008
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Media;
using System.IO;
using System.Windows.Forms;

namespace Etaris.Core
{
    /// <summary>
    /// Statussen van de muziekspeler.
    /// </summary>
    internal enum MediaPlayerStates
    {
        /// <summary>
        /// De MusicPlayer speelt muziek af.
        /// </summary>
        Playing,
        /// <summary>
        /// De muziek is gepauzeerd.
        /// </summary>
        Paused,
        /// <summary>
        /// De muziek is afgelopen of gestopt.
        /// </summary>
        Stopped,
        /// <summary>
        /// Toestand van de MusicPlayer voordat de muziek is gestart.
        /// </summary>
        Empty
    }

    /// <summary>
    /// Implementeert een simpele muziekspeler.
    /// </summary>
    public class MusicPlayer
    {
        #region Velden
        /// <summary>
        /// De lijst met muziekbestanden die afgespeeld worden.
        /// </summary>
        private List<Uri> playList = new List<Uri>();

        /// <summary>
        /// De daadwerkelijke mediaspeler.
        /// </summary>
        private MediaPlayer mp = new MediaPlayer();

        /// <summary>
        /// De toestand waarin de mediaspeler zich momenteel bevindt.
        /// </summary>
        private MediaPlayerStates mpState = MediaPlayerStates.Empty;

        /// <summary>
        /// De random-getallen generator. Hiermee wordt het volgende nummer bepaald.
        /// </summary>
        private Random rng = new Random();

        /// <summary>
        /// De index in <i>playList</i> van het muziekbestand dat momenteel afgespeeld wordt.
        /// </summary>
        private int currentFile;
        #endregion

        /// <summary>
        /// Constructor. Initialiseert de klasse met een gegeven afspeellijst.
        /// </summary>
        /// <param name="files">Lijst van muziekbestanden voor de muziekspeler om te spelen.</param>
        public MusicPlayer(List<String> files)
        {
            for (int i = 0; i < files.Count; i++)
                if (File.Exists(Path.GetFullPath(files[i])))
                    playList.Add(new Uri(Path.GetFullPath(files[i])));

            if(playList.Count != 0) currentFile = rng.Next(0, playList.Count-1);

            mp.MediaEnded += new EventHandler(MediaEnded);
            mp.MediaFailed += new EventHandler<ExceptionEventArgs>(MediaFailed);
        }

        /// <summary>
        /// Speelt een willekeurig muziekbestand (maar niet het vorige).
        /// </summary>
        public void Play()
        {
            //Init.
            if (playList.Count == 0) return;
            
            //Kiest een ander nummer als er andere nummers zijn
            if (playList.Count != 1) currentFile = (currentFile + rng.Next(1, playList.Count - 1)) % playList.Count;

            //Speelt geselectuurde file
            mp.Stop(); mp.Open(playList[currentFile]); mp.Play();
            mpState = MediaPlayerStates.Playing;
        }

        /// <summary>
        /// Stopt het spelen van muziek.
        /// </summary>
        public void Stop()
        {
            mp.Stop();
            mpState = MediaPlayerStates.Stopped;
        }

        /// <summary>
        /// Pauzeert en hervat de muziekspeler.
        /// </summary>
        public void Pause()
        {
            if (mpState == MediaPlayerStates.Playing)
            {
                mp.Pause();
                mpState = MediaPlayerStates.Paused;
            }
            else
            {
                mp.Play();
                mpState = MediaPlayerStates.Playing;
            }
        }

        /// <summary>
        /// Probeert een ander muziekbestand af te spelen als het huidige is afgelopen.
        /// </summary>
        private void MediaEnded(object sender, EventArgs e)
        {
            Play();
        }

        /// <summary>
        /// Verwijdert een muziekbestand dat een fout gegenereerd heeft en probeert een ander muziekbestand af te spelen.
        /// </summary>
        private void MediaFailed(object sender, ExceptionEventArgs e)
        {
            playList.RemoveAt(currentFile);
            if (playList.Count == 1) currentFile = 0;
            Play();
        }
    }
}
