﻿/*** Name:      Piece.cs
 *** Created:   10/25/2008, by: Hilkens, RGM
 *** Modified:  12/14/2008, by: van Vliembergen, RWL
 *** Source:    Definieert een speelstuk.
 *** Copyright: (c) Gr2 Dev 2008
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using sysui;

namespace Etaris.Fields
{
    /// <summary>
    /// Definieert een speelstuk. Een speelstuk moet consistent zijn. De graaf van aaneenliggende blokken moet 'connected' zijn (consistentie).
    /// </summary>
    /// <remarks>
    /// Met Top en Left worden altijd plaatsen in het speelveld bedoeld (zoals in Grid).
    /// Bij het roteren wordt gebruik gemaakt van een coördinatensysteem met een tweemaal hogere
    /// resolutie dan het speelveld. Een coördinaat in dit systeem is valide, wanneer het precies
    /// met een hoek- of middelpunt van een cel in het speelveld overeenkomt.
    /// Dan geeft roteren over 90 graden van een blok om dit punt namelijk weer een geldig punt.
    /// </remarks>
    public partial class Piece : BasePiece<Block>
    {
        /// <summary>
        /// Het speelveld waarin dit speelstuk zich bevindt.
        /// </summary>
        PlayingField field;

        /// <summary>
        /// Een willekeurige waarde die vrijelijk gebruikt kan worden door andere klassen.
        /// </summary>
        int tag;

        /// <summary>
        /// Het massamiddelpunt van dit speelstuk.
        /// </summary>
        MassCentre massCentre = new MassCentre();

        #region "Properties"
        /// <summary>
        /// Het bij dit speelstuk behorende speelveld.
        /// </summary>
        public PlayingField Field
        {
            get { return field; }
        }

        /// <summary>
        /// Een willekeurige waarde die vrijelijk gebruikt kan worden door andere klassen.
        /// </summary>
        /// <remarks>
        /// Deze waarde wordt gebruikt door PlayingField.RemoveRows.
        /// </remarks>
        internal int Tag
        {
            get { return tag; }
            set { tag = value; }
        }
        #endregion

        /// <summary>
        /// Constructor. Initialiseert de klasse.
        /// </summary>
        /// <param name="playingField">Speelveld behorende bij dit speelstuk.</param>
        /// <param name="color">Hoofdkleur van dit speelstuk.</param>
        public Piece(PlayingField playingField, Color color)
            : base(color)
        {
            massCentre.IsCalculated = false;
            field = playingField;
            field.Pieces.AddLast(this);
            SetPaintingSizes(field.BlockSize, field.ShadingSize);
        }

        /// <summary>
        /// Verplaatst dit speelstuk wanneer mogelijk één stap in de gegeven richting.
        /// </summary>
        /// <param name="direction">De richting waarin bewogen moet worden.</param>
        /// <returns><i>true</i> als het speelstuk is bewogen in de aangegeven richting, anderzijds <i>false</i>.</returns>
        public bool Move(Direction direction)
        {
            //Controleer of verplaatsing mogelijk is.
            if (this.CanMove(direction))
            {
                //Wis oude afbeelding.
                ClearDrawing(field.GraphicsContext);

                //Verplaats het speelstuk.
                switch (direction)
                {
                    case Direction.Left: //Naar links bewegen.
                        foreach (Block block in Blocks) block.SetNewPosition(block.Left - 1, block.Top);
                        massCentre.X -= 2;
                        break;

                    case Direction.Right: //Naar rechts bewegen.
                        foreach (Block block in Blocks) block.SetNewPosition(block.Left + 1, block.Top);
                        massCentre.X += 2;
                        break;

                    case Direction.Down: //Naar beneden bewegen.
                        foreach (Block block in Blocks) block.SetNewPosition(block.Left, block.Top + 1);  //afstand tot Top wordt groter.
                        massCentre.Y += 2;
                        break;

                    case Direction.Up: //Naar boven bewegen.
                        foreach (Block block in Blocks) block.SetNewPosition(block.Left, block.Top - 1); //afstand tot Top wordt kleiner.
                        massCentre.Y -= 2;
                        break;
                }

                //Teken speelstuk.
                Paint(field.GraphicsContext);
                return true;
            }
            else return false;
        }

        /// <summary>
        /// Verplaatst dit speelstuk wanneer mogelijk maximaal in de gegeven richting.
        /// </summary>
        /// <param name="direction">De richting waarin bewogen moet worden.</param>
        /// <returns><i>true</i> als het speelstuk minstens één hokje is bewogen in de aangegeven richting, anderzijds <i>false</i>.</returns>
        public bool MoveMaximal(Direction direction)
        {
            //Init.
            bool ret = false;
            bool moved;

            do
            {
                moved = Move(direction); //Beweeg het speelstuk 1 stap in de aangegeven richting, indien dit kan
                ret = ret || moved;
            } while (moved);

            //Finish.
            return ret;
        }

        /// <summary>
        /// Draait het speelstuk indien mogelijk 90 graden met de klok mee.
        /// </summary>
        /// <returns><i>true</i> indien het roteren gelukt is, anders <i>false</i>.</returns>
        public bool Rotate()
        {
            //Init.
            int xOld; int xNew;
            int yOld; int yNew;
            BoundingBox box2, box1 = new BoundingBox(); //Dit kan eventueel in PIECE opgeslagen worden (caching).

            //Wis oude afbeelding.
            ClearDrawing(field.GraphicsContext);

            //Bereken massamiddelpunt als dit nog niet gedaan is.
            if (!massCentre.IsCalculated) CalculateMassCentre();

            //Controleer of draaien mogelijk is. Met een struct Vector kan het een en ander netter. Echter, dan moet veel code gewijzigd worden.
            //In dat geval zijn twee extra regels in deze functie beter. Heb je vragen, neem dan EERST contact met Rick op!!!
            box1.Calculate(this);
            box2 = box1;

            for (int i = 0; i < 2; i++) //Roteer box2.
            {
                if (i == 0) { xOld = box2.x1 * 2; yOld = box2.y1 * 2; } else { xOld = box2.x2 * 2; yOld = box2.y2 * 2; }

                xNew = (-(yOld - massCentre.Y)) + massCentre.X;
                yNew = ((xOld - massCentre.X)) + massCentre.Y;

                if (i == 0) { box2.x1 = xNew / 2; box2.y1 = yNew / 2; } else { box2.x2 = xNew / 2; box2.y2 = yNew / 2; ; }
            }

            box2.MakeValid(); box1.Union(box2);
            if (!box1.IsGridFree(this)) { Paint(field.GraphicsContext); return false; }

            //Roteer (wiskundig equivalent aan de rotaties hierboven).
            foreach (Block block in blocks)
            {
                xNew = (-((block.Top * 2) - massCentre.Y)) + massCentre.X;
                yNew = (((block.Left * 2) - massCentre.X)) + massCentre.Y;
                block.SetNewPosition(xNew / 2, yNew / 2);
            }

            //Finish.
            Paint(field.GraphicsContext);
            return true;
        }

        /// <summary>
        /// Voegt een blok toe (houdt geen rekening met consistentie).
        /// </summary>
        /// <param name="left">Kolom in speelveld (beginnend vanaf 0). De eenheid is een blokbreedte.</param>
        /// <param name="top">Rij in speelveld (beginnend vanaf 0). De eenheid is een blokhoogte.</param>
        public void AddBlock(int left, int top)
        {
            Block b = new Block(this, left, top);
            Blocks.Add(b);
        }

        /// <summary>
        /// Verwijdert een aantal blokken en splitst het speelstuk op indien nodig.
        /// </summary>
        /// <param name="blocks">Lijst met de blokken die verwijderd moeten worden.</param>
        internal void RemoveBlocks(List<Block> blocks)
        {
            //Init.
            Block tblock; //tijdelijk blok
            Piece tpiece; //tijdelijk speelstuk
            List<Block> tlist = new List<Block>(); //tijdelijke lijst van blokken
            int left, top; //Coordinaten van een blok

            int[] dx = new int[] { 0, 0, -1, 1 };
            int[] dy = new int[] { -1, 1, 0, 0 };

            //Wis oude afbeelding.
            ClearDrawing(field.GraphicsContext);

            //Verwijdert de nodige blokken uit het speelstuk:
            foreach (Block b in blocks)
            {
                field.Grid[b.Left, b.Top] = null;
                field.RowItemCount[b.Top]--;
                Blocks.Remove(b);
            }

            //Kijkt of er blokken over zijn en verplaatst die blokken naar nieuwe speelstukken
            while (Blocks.Count != 0)
            {
                //Haalt eerste block uit piece
                tblock = Blocks[0];

                //Creëert nieuw piece
                tpiece = new Piece(field, Color);

                //Zet blok is lijst
                tlist.Add(tblock);

                //Gooi blok uit oude piece en voeg toe aan nieuwe
                Blocks.Remove(tblock);
                tpiece.Blocks.Add(tblock);
                tblock.Piece = tpiece;

                //Loop om rest van nieuwe piece te vinden
                while (tlist.Count != 0)
                {
                    //Haal eerste blok uit list
                    tblock = tlist[0]; tlist.RemoveAt(0);

                    //Haal coördinaten op
                    left = tblock.Left; top = tblock.Top;

                    //Buren afgaan om te kijken of in deze piece
                    for (int i = 0; i < 4; i++)
                        if (field.GetPiece(left + dx[i], top + dy[i]) == this)
                        {
                            tblock = field.Grid[left + dx[i], top + dy[i]];

                            //Voeg blok toe aan lijst
                            tlist.Add(tblock);

                            //Verwijder dit blok uit het oude speelstuk en plaats het in het nieuwe speelstuk
                            Blocks.Remove(tblock);
                            tpiece.Blocks.Add(tblock);
                            tblock.Piece = tpiece;
                        }
                }

                //Teken dit nieuwe stuk
                tpiece.Paint(field.GraphicsContext);
            }

            //Finish.
            field.Pieces.Remove(this); //Verwijder dit lege speelstuk
        }

        /// <summary>
        /// Controleert of een speelstuk in een bepaalde richting kan bewegen.
        /// </summary>
        /// <param name="direction">De richting waarin getest wordt of bewogen kan worden.</param>
        /// <returns>Geeft <i>true</i> als het speelstuk kan bewegen in de aangegeven richting, anders <i>false</i>.</returns>
        private bool CanMove(Direction direction)
        {
            foreach (Block block in Blocks)
                if (!block.CanMove(direction))
                    return false; //Op het moment dat 1 blok niet kan bewegen, mag het hele speelstuk niet bewegen.

            return true; //Geen reden gevonden waarom het speelstuk niet zou kunnen bewegen.
        }

        /// <summary>
        /// Berekent het massamiddelpunt. Bij een niet valide massamiddelpunt
        /// worden de coördinaten in de richting van het midden van het speelstuk
        /// afgerond zodat deze wel valide zijn.
        /// </summary>
        private void CalculateMassCentre()
        {
            //Init.
            int x = 0, y = 0;
            int xMin, xMax, yMin, yMax;
            int xTemp, yTemp;

            //Initialiseer xMin t/m yMax.
            xMin = Blocks[0].Left * 2;
            yMin = Blocks[0].Top * 2;
            xMax = xMin;
            yMax = yMin;

            //Bereken zwaartepunt.
            foreach (Block block in Blocks)
            {
                xTemp = block.Left * 2;
                yTemp = block.Top * 2;
                if (xTemp < xMin) xMin = xTemp; else if (xTemp > xMax) xMax = xTemp;
                if (yTemp < yMin) yMin = yTemp; else if (yTemp > yMax) yMax = yTemp;
                x += xTemp; y += yTemp;
            }

            //Finish.
            CalculateValidMassCentre((xMax + xMin) / 2, (yMax + yMin) / 2, (int)Math.Round((double)x / (double)(Blocks.Count), 0), (int)Math.Round((double)y / (double)(Blocks.Count), 0));
        }

        /// <summary>
        /// Controleert een massamiddelpunt op validiteit en maakt het zonodig valide.
        /// </summary>
        /// <remarks>
        /// Een massamiddelpunt is niet valide indien 90 graden eromheen draaien de blokken op ongeldige posities zet.
        /// </remarks>
        /// <param name="cX">X-coördinaat van middelpunt.</param>
        /// <param name="cY">Y-coördinaat van middelpunt.</param>
        /// <param name="mX">X-coördinaat van massamiddelpunt.</param>
        /// <param name="mY">Y-coördinaat van massamiddelpunt.</param>
        private void CalculateValidMassCentre(int cX, int cY, int mX, int mY)
        {
            //Init.
            bool doX = (Math.Abs(cX - mX) < Math.Abs(cY - mY));

            //Controleer of mX is oneven en mY is even of omgekeerd.
            if ((mX % 2 == 0) ^ (mY % 2 == 0))
            {
                if (doX && mX != cX || (!doX && mY == cY))
                {
                    if (mX < cX) mX++; else mX--; //Verander x
                }
                else if (!doX && mY != cY || (doX && mX == cX))
                {
                    if (mY < cY) mY++; else mY--; //Verander y
                }
                else mX++;   //Geen voorkeur, willkeurig
            }

            //Finish.
            massCentre.X = mX;
            massCentre.Y = mY;
            massCentre.IsCalculated = true;
        }
    }

    /// <summary>
    /// Definieert een bewegingsrichting in het speelveld.
    /// </summary>
    public enum Direction
    {
        /// <summary>
        /// Bewegingsrichting langs de negatieve Y-as.
        /// </summary>
        Up,
        /// <summary>
        /// Bewegingsrichting langs de positieve Y-as.
        /// </summary>
        Down,
        /// <summary>
        /// Bewegingsrichting langs negatieve X-as.
        /// </summary>
        Left,
        /// <summary>
        /// Bewegingsrichting langs positieve X-as.
        /// </summary>
        Right
    }

    /// <summary>
    /// Definieert zijdes van een blok.
    /// </summary>
    internal enum Sides
    {
        /// <summary>
        /// Bovenste zijde van een blok.
        /// </summary>
        Top,
        /// <summary>
        /// Onderste zijde van een blok.
        /// </summary>
        Bottom,
        /// <summary>
        /// Linker zijde van een blok.
        /// </summary>
        Left,
        /// <summary>
        /// Rechter zijde van een blok.
        /// </summary>
        Right
    }
}
